from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.validators import attributes_validator
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.modules.resources._machine import Machine

required_attributes = [
    'title',
    'ssh_ip',
    'ssh_port',
    'ssh_user'
]

default_attributes = {
    "ram": 1.0,
    "cpu_cores": 1.0
}


class MachinesClient:

    def __init__(self, organization):
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.MACHINES_BASE.format(scope["organization"])

    def list(self, sort="-id"):
        """
        List all the machines in the organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields Machine objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Machine,
            sort=sort,
        )

    def get(self, slug):
        """
        Retrieves a machine by the given slug
        @param slug: The slug of the requested machine
        @return: Machine object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.MACHINE_GET_FAULTY_SLUG)

        return Machine(context=self._context, slug=slug)

    def create(
        self,
        title,
        ssh_user,
        ssh_ip,
        ssh_port,
        ssh_key=None,
        ssh_password=None,
        **kwargs
    ):

        # --- #
        # A machine will generate a singular template using these values,
        # so we handle creation here and not in the machine templates client
        # --- #

        """
        @param title: Name of the new machine
        @param ssh_user: Credentials for the new machine
        @param ssh_ip: Host of the new machine
        @param ssh_port: Port of the new machine
        @param ssh_key: a ssh key file for the new machine(path to file relative to working dir)
        @param ssh_password: a password for the new machine
        @param kwargs: A list of optional attributes for creation
            cpu_cores: CPU cores for the new machine
            ram: memory for the new machine
            gpu: Does the new machine has a GPU
            gpu_model: What GPU the new machine has
            gpu_cores: GPU Cores for the new machine
            gpu_ram: GPU memory for the new machine
        @return: The newly created machine
        """

        attributes = {
            "title": title,
            "ssh_user": ssh_user,
            "ssh_ip": ssh_ip,
            "ssh_port": ssh_port,
            **default_attributes,
            **kwargs,
        }

        if ssh_key:
            # todo: Add ssh_key file support
            pass
        else:
            attributes["ssh_password"] = ssh_password

        attributes_validator(
            Machine.machine_attributes,
            attributes,
            required_attributes
        )

        # Instance specific validations
        if not ssh_password and not ssh_key:
            raise CnvrgArgumentsError(error_messages.MACHINE_NO_CREDENTIALS.format('create'))

        if not kwargs.get("gpu") and (kwargs.get("gpu_model") or kwargs.get('gpu_cores') or kwargs.get('gpu_ram')):
            raise CnvrgArgumentsError(error_messages.MACHINE_GPU_VALUES_WITHOUT_GPU)

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type='machine', attributes=attributes),
        )
        slug = response.attributes['slug']
        return Machine(context=self._context, slug=slug, attributes=response.attributes)



from cnvrgv2.config import routes
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.json_api_format import JAF


class DatasetCommit(DynamicAttributes):

    available_attributes = {
        "sha1": str,
        "source": str,
        "message": str,
        "created_at": str,
        "use_cached": str,
    }

    def __init__(self, context=None, slug=None, attributes=None):

        self._context = Context(context=context)

        # Set current context scope to current dataset
        if slug:
            self._context.set_scope(SCOPE.COMMIT, slug)

        self.scope = self._context.get_scope(SCOPE.COMMIT)

        self._proxy = Proxy(context=self._context)
        self._route = routes.DATASET_COMMIT_BASE.format(self.scope["organization"], self.scope["dataset"],
                                                        self.scope["commit"])
        self._attributes = attributes or {}

    def cache_commit(self, external_disk_title):
        """
        Caches the current commit onto an external disk, for quick access
        @param external_disk_title: title of the external disk to cache commit on
        @return: None
        """
        cache_commit_route = routes.WORKFLOW_STANDALONE_BASE.format(self.scope["organization"])

        attributes = {
            "commit_hash": self.sha1,
            "dataset_slug": self.scope["dataset"],
            "external_disk_title": external_disk_title,
            "task": "cache"
        }

        self._proxy.call_api(
            route=cache_commit_route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="DatasetWarmup", attributes=attributes)
        )
        # TODO: Should return something? Update docstring

    def clear_cached_commit(self, external_disk_title):
        """
        Removes the current cached commit from the external disk.
        @param external_disk_title: title of the external disk to remove cached commit from
        @return: None
        """
        clear_cached_commit_route = routes.WORKFLOW_STANDALONE_BASE.format(self.scope["organization"])

        attributes = {
            "commit_hash": self.sha1,
            "dataset_slug": self.scope["dataset"],
            "external_disk_title": external_disk_title,
            "task": "clear"
        }

        self._proxy.call_api(
            route=clear_cached_commit_route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="DatasetWarmup", attributes=attributes)
        )
        # TODO: Should return something? Update docstring



import re
import base64
import mimetypes
import os


def urljoin(*args):
    """
    Joins given arguments into an url. Trailing but not leading slashes are
    stripped for each argument.
    @param args: strings to join into a url
    @return: Valid url string
    """

    joined_url = "/".join(map(lambda x: str(x).strip('/'), args))
    return format_url(joined_url)


def format_url(url):
    # remove double slashes from the url to support oauth proxy
    return re.sub(r'[\/]{2,}', "/", url).replace("http:/", "http://").replace("https:/", "https://")


def encode_base64(file_path, encoding_type="utf-8"):
    """
    Base 64 converter
    @param file_path: A file path
    @param encoding_type: Type of encoding, default utf-8
    @return: An object including filename, type, size and base64 content like a File spec
    """
    if file_path:
        file = open(file_path, "rb")
        binary_file = file.read()
        file_type = mimetypes.guess_type(file_path)[0]

        return {
            "filename": os.path.basename(file_path),
            "type": file_type,
            "size": os.path.getsize(file_path),
            "content": "data:" + file_type + ";base64," + (base64.b64encode(binary_file)).decode(encoding_type)
        }



from cnvrgv2.config import routes
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.errors import CnvrgError
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.modules.resources.templates.kube_templates_client import KubeTemplatesClient


class Cluster(DynamicAttributes):
    available_attributes = {
        # TODO: Add all relevant attributes
        "slug": str,
        "title": str,
        "status": str,
        "is_default": bool
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current machine
        if slug:
            self._context.set_scope(SCOPE.RESOURCE, slug)

        scope = self._context.get_scope(SCOPE.RESOURCE)

        self._proxy = Proxy(context=self._context)
        self._route = routes.CLUSTER_BASE.format(scope["organization"], scope["resource"])
        self._attributes = attributes or {}
        self.slug = scope["resource"]
        self._init_clients()

    def save(self):
        pass

    def delete(self):
        """
        Deletes the current spark driver
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)

    def _init_clients(self):
        """
        Sets up the clients that are exposed to user user via machine
        @return: Void
        """
        try:
            self.templates = KubeTemplatesClient(context=self._context)
        except CnvrgError:
            # TODO: Surpress exceptions here
            pass



import io
import os
import shlex
import subprocess
import sys
import threading
import re
from contextlib import redirect_stdout, redirect_stderr
import inspect

from cnvrgv2.config.error_messages import MISSING_EXPERIMENT_ARGUMENT, NOT_A_DATASET_LIST_OBJECT, NOT_A_VOLUME_OBJECT, \
    EXPERIMENT_CANNOT_BE_A_FUNCTION
from cnvrgv2.context import SCOPE
from cnvrgv2.config import routes
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.base.workflows_base import WorkflowsBase
from cnvrgv2.modules.volumes.volume import Volume
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.modules.workflows import Experiment
from cnvrgv2.utils.log_utils import LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR, log_buffer
from cnvrgv2.utils.validators import validate_types_in_list
from cnvrgv2.utils.env_helper import ENV_KEYS


class ExperimentsClient(WorkflowsBase):
    def __init__(self, project):
        super().__init__(Experiment, "Experiment", project._context)

        scope = self._context.get_scope(SCOPE.PROJECT)
        self.project = project
        self._route = routes.EXPERIMENTS_BASE.format(scope["organization"], scope["project"])

    def init(self, title=None, **kwargs):
        """
        Creates an experiment that doesn't run anything locally or on remote
        @param title: Title of the experiment
        @param kwargs: Additional arguments
        @return: The new experiment object
        """
        kwargs = {
            "clean_experiment": True,
            **kwargs
        }
        return self.create(title=title, local=True, sync_before=False, sync_after=False, **kwargs)

    def create(
        self,
        title=None,
        templates=None,
        local=False,
        command=None,
        local_arguments=None,
        datasets=None,
        volume=None,
        sync_before=True,
        sync_after=True,
        **kwargs
    ):
        """
        Create and run a new experiment
        @param title: Name of the experiment
        @param templates: Templates to run the experiment on
        @param local: Boolean. Run the experiment locally
        @param command: The command to run. can be a function in case of a local experiment
        @param local_arguments: If local experiment and command is a function,
               local_arguments is a dictionary of the arguments to pass to the experiment's function
        @param datasets: list of datasets to use in the experiment.
        @param volume: A volume to attach to this experiment. (type: Volume object).
        @param args: optional arguments
        @param sync_before: Boolean. sync environment before running the experiment
        @param sync_after: Boolean. sync environment after the experiment finished
        @param kwargs: Dictionary. Rest of optional attributes for creation
            image: Image object to create experiment with
            git_branch: The branch to pull files from for the experiment, in case project is git project
            git_commit: The commit to pull files from for the experiment, in case project is git project
            queue: Name of the queue to run this job on
        @return: The experiment's object
        """
        m_kwargs = {
            "command": command,
            "local": local,
            **kwargs
        }

        if volume and not isinstance(volume, Volume):
            raise CnvrgArgumentsError(NOT_A_VOLUME_OBJECT)
        elif volume:
            m_kwargs["external_disk_slug"] = volume.slug

        if datasets and not validate_types_in_list(datasets, Dataset):
            raise CnvrgArgumentsError(NOT_A_DATASET_LIST_OBJECT)
        elif datasets:
            m_kwargs["job_datasets"] = [ds.as_request_params() for ds in datasets]

        if callable(command):
            if local:
                # if command is function, run it as a function and not as a os command. Update kwargs appropriately
                func_args = inspect.signature(command).parameters.keys()
                if "experiment" not in func_args:
                    raise CnvrgArgumentsError(MISSING_EXPERIMENT_ARGUMENT)
                m_kwargs["command"] = "func: {func_name}".format(func_name=command.__name__)
            else:
                raise CnvrgArgumentsError(EXPERIMENT_CANNOT_BE_A_FUNCTION)

        if sync_before and not self.project.git:
            self.project.sync()

        m_kwargs['job_parent_id'] = os.environ.get(ENV_KEYS['current_job_id'])

        experiment = super().create(title, templates, **m_kwargs)

        if local and not m_kwargs.get("clean_experiment", False):
            local_arguments = local_arguments or {}
            exit_status = self._prepare_local_experiment(
                experiment=experiment,
                executable=command,
                working_dir=m_kwargs.get("cwd", None),
                arguments={"experiment": experiment, **local_arguments},
            )

            if sync_after:
                self.project.sync()

            experiment.finish(exit_status)

        return experiment

    def _prepare_local_experiment(self, experiment, executable, working_dir, arguments: dict = None):
        """
        Prepares the local experiment before running and runs the experiment
        @param experiment: The experiment's object
        @param executable: The function to be run locally
        @param working_dir: The working dir
        @param arguments: In case user sent a function: Additional arguments to send to the user's function
        @return: execution's exit status
        """
        if callable(executable):
            with redirect_stdout(io.StringIO()) as stdout, redirect_stderr(io.StringIO()) as stderr:
                return self._run_local_experiment(
                    experiment=experiment,
                    arguments=arguments,
                    stdout=stdout,
                    stderr=stderr,
                    exp_func=executable,
                )
        else:
            env = experiment.as_env()
            process = self._run_subprocess(cmd=executable, cwd=working_dir, env=env)
            stdout = process.stdout
            stderr = process.stderr
            return self._run_local_experiment(
                experiment=experiment,
                stdout=stdout,
                stderr=stderr,
                command_proc=process
            )

    @staticmethod
    def _run_subprocess(cmd, cwd=None, env=None):
        """
        Opens a subprocess that runs the user's command locally
        @param cmd: The command to run
        @param cwd: The working dir to work in
        @param env: Environment parameters
        @return: The subprocess running the command
        """
        env = dict({**os.environ, **(env or {}), **{'PYTHONUNBUFFERED': "1"}})
        cmd = re.sub(r"^(python3?)", r"{exe} -u".format(exe=sys.executable), cmd)
        cmd = shlex.split(cmd)

        params = {
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
            'close_fds': False,
            'bufsize': 1,
        }
        if cwd:
            params['cwd'] = r"{}".format(cwd)
        if env:
            params["env"] = env
        return subprocess.Popen(cmd, **params)

    @staticmethod
    def _run_local_experiment(experiment, stdout, stderr, arguments=None, exp_func=None, command_proc=None):
        """
        Runs local experiment and tracks logs written during the experiment
        @param experiment: The experiment's object
        @param stdout: The standard output logs are written to
        @param stderr: The standard errors output logs are written to
        @param arguments: In case user sent a function: Additional arguments to send to the user's function
        @param exp_func: The user's function to run, in case command was a function
        @param command_proc: The function to run the user's command, in case a command was sent
        @return: exit status of the experiment's execution
        """

        exit_status = None

        t_out = threading.Thread(target=log_buffer, kwargs={
            "buffer": stdout,
            "workflow": experiment,
            "log_type": LOGS_TYPE_OUTPUT
        })

        t_err = threading.Thread(target=log_buffer, kwargs={
            "buffer": stderr,
            "workflow": experiment,
            "log_type": LOGS_TYPE_ERROR
        })

        t_out.start()
        t_err.start()

        try:
            if exp_func:
                exit_status = exp_func(**arguments)
                result_log = exp_func.__name__ + " returned " + str(exit_status) + " as exit status."
                stdout.write(result_log)
                stdout.write('\0')
                stderr.write('\0')

        except Exception as e:
            exit_status = 1
            experiment.write_logs(str(e), LOGS_TYPE_ERROR)
        except KeyboardInterrupt:
            exit_status = -100
            experiment.write_logs("Keyboard Interrupt", LOGS_TYPE_ERROR)

        t_out.join()
        t_err.join()

        stdout.close()
        stderr.close()

        if command_proc:
            command_proc.wait()
            exit_status = command_proc.returncode

        # If user's function finished without failing and didn't return a status, we say it succeeded
        return exit_status or 0



from cnvrgv2 import cnvrg
from cnvrgv2.config import error_messages, routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.project import Project
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin


class ProjectsClient:
    def __init__(self, organization):
        if type(organization) is not cnvrg.Cnvrg:
            raise CnvrgArgumentsError(error_messages.ARGUMENT_BAD_TYPE.format(cnvrg.Cnvrg, type(organization)))
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.PROJECTS_BASE.format(scope["organization"])

    def list(self, sort="-id"):
        """
        List all projects in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields project objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Project,
            sort=sort
        )

    def get(self, slug):
        """
        Retrieves a project by the given slug
        @param slug: The slug of the requested project
        @return: Project object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.PROJECT_GET_FAULTY_SLUG)

        return Project(context=self._context, slug=slug)

    def create(self, name):
        """
        Creates a new project with the given name
        @param name: name of the new project
        @return: the newly created project object
        """
        if not name or not isinstance(name, str):
            raise CnvrgArgumentsError(error_messages.PROJECT_CREATE_FAULTY_NAME)

        attributes = {"title": name}
        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="project", attributes=attributes)
        )

        slug = response.attributes['slug']
        return Project(context=self._context, slug=slug)

    def delete(self, slugs):
        """
        Deleting multiple projects
        @param slugs: List of project slugs to be deleted
        @return: None
        """
        delete_url = urljoin(self._route, routes.PROJECT_BULK_DELETE_SUFFIX)

        attributes = {
            "project_slugs": slugs
        }

        self._proxy.call_api(
            route=delete_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="Project", attributes=attributes)
        )



from cnvrgv2.modules.images.image import Image
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.context import Context
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.url_utils import urljoin


class WorkflowsBase:
    def __init__(self, object, type, context=None):
        """
        Base class for all kind of workflows
        @param object: The current workflow object type
        @param type: String represents the resource type to be created
        @param context: The context object
        """
        self._object = object
        self._type = type
        self._context = Context(context=context)
        self._proxy = Proxy(context=self._context)

    def list(self, sort="-id"):
        """
        List all workflows of a specific type in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields project objects
        """
        return api_list_generator(
            context=self._context,
            route=urljoin(self._route, routes.WORKFLOW_LIST_TYPE_SUFFIX.format(self._type)),
            object=self._object,
            sort=sort
        )

    def get(self, slug):
        """
        Retrieves a project by the given slug
        @param slug: The slug of the requested project
        @return: Project object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.PROJECT_GET_FAULTY_SLUG)

        return self._object(context=self._context, slug=slug)

    def create(self, title, templates, *args, **kwargs):
        """
        Creates a new workflow
        @param title: Name of the workflow
        @param templates: List of template names to be used
        @param args: optional arguments
        @param kwargs: Dictionary. Rest of optional attributes for creation
            image: Image object to create workflow with
            queue: Name of the queue to run this job on
        @return: The newly created workflow object
        """

        if isinstance(kwargs.get("image"), Image):
            kwargs["image_slug"] = kwargs.get("image").slug
            del kwargs["image"]

        if kwargs.get("queue", False):
            kwargs["queue_name"] = kwargs.get("queue", False)
            del kwargs["queue"]

        attributes = {
            "title": title,
            "template_names": templates,
            **kwargs
        }

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

        slug = response.attributes['slug']
        return self._object(context=self._context, slug=slug)

    def delete(self, slugs):
        """
        Deleting multiple workflows
        @param slugs: List of workflow slugs to be deleted
        @return: None
        """
        delete_url = urljoin(self._route)

        attributes = {
            "workflow_slugs": slugs
        }

        self._proxy.call_api(
            route=delete_url,
            http_method=HTTP.DELETE,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def stop(self, slugs, sync=False):
        """
        Stopping multiple workflows
        @param slugs: List of workflow slugs to be stopped
        @param sync: Sync workflow before stopping. Default: false
        @return: None
        """
        stop_url = urljoin(self._route, routes.WORKFLOW_STOP_SUFFIX)

        attributes = {
            "workflow_slugs": slugs,
            "sync": sync
        }

        self._proxy.call_api(
            route=stop_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def sync_remote(self, slugs, commit_msg=None):
        """
        Syncing multiple workflows
        @param slugs: List of workflow slugs to be synced
        @param commit_msg: Commit message to add
        @return: None
        """
        sync_url = urljoin(self._route, routes.WORKFLOW_SYNC_SUFFIX)

        attributes = {
            "workflow_slugs": slugs
        }

        if commit_msg:
            attributes["commit_msg"] = commit_msg

        self._proxy.call_api(
            route=sync_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )



from cnvrgv2.config import routes, error_messages
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.commits.dataset_commit import DatasetCommit
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.modules.base.data_owner import DataOwner
from cnvrgv2.modules.queries.queries_client import QueriesClient
from cnvrgv2.errors import CnvrgError
from cnvrgv2.config.error_messages import LOCAL_COMMIT_DOESNT_EXIST_ERROR
import os
from cnvrgv2.config import Config


class SyncType:
    ALL = "sync_all"
    FLATTEN = "sync_flatten"


class Dataset(DataOwner):

    # TODO: add all relevant attributes.
    available_attributes = {
        "slug": str,
        "size": int,
        "title": str,
        "members": list,
        "datatype": str,
        "category": str,
        "description": str,
        "num_files": int,
        "last_commit": str,
        "is_public": bool
    }

    def __init__(self, context=None, slug=None, attributes=None):
        # Init data attributes
        super().__init__()

        self._context = Context(context=context)

        # Set current context scope to current dataset
        if slug:
            self._context.set_scope(SCOPE.DATASET, slug)

        self.scope = self._context.get_scope(SCOPE.DATASET)

        self._proxy = Proxy(context=self._context)
        self._route = routes.DATASET_BASE.format(self.scope["organization"], self.scope["dataset"])
        self._attributes = attributes or {}
        self.slug = self.scope["dataset"]
        self.use_cached = False
        self.sync_type = SyncType.ALL

        self._init_clients()

    def get_commit(self, sha_1):
        if not sha_1 or not isinstance(sha_1, str):
            raise CnvrgArgumentsError(error_messages.COMMIT_FAULTY_SHA1)

        return DatasetCommit(context=self._context, slug=sha_1)

    def list_commits(self, sort="-id"):
        """
        List all commits in a specific dataset
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields commit objects
        """
        list_commits_url = routes.DATASET_COMMITS_BASE.format(self.scope["organization"], self.scope["dataset"])

        return api_list_generator(
            context=self._context,
            route=list_commits_url,
            object=DatasetCommit,
            sort=sort,
            identifier="sha1"
        )

    def delete(self):
        """
        Deletes the current dataset
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)

    def _validate_config_ownership(self):
        return self.slug == self._config.dataset_slug

    def as_request_params(self):
        return {
            "slug": self.slug,
            "sync_type": self.sync_type,
            "commit": self.local_commit,
            "query_slug": "",  # TODO: add query slug
            "use_cached": self.use_cached,
        }

    def _init_clients(self):
        self.queries = QueriesClient(self._context)

    def save_config(self, local_config_path=None):
        """
        Saves the dataset configuration in the local folder
        @return: None
        """

        if not self.local_commit:
            raise CnvrgError(LOCAL_COMMIT_DOESNT_EXIST_ERROR)
        self._config.update(local_config_path=local_config_path, **{
            "dataset_slug": self.slug,
            "organization": self._context.organization,
            "commit_sha1": self.local_commit,
        })

    def verify(self):
        """
        Checks if the dataset has been created successfully (if the config file has been created)
        @return: returns True if the config file has been created
        """
        cwd = os.getcwd()
        if not os.path.isdir(self.slug):
            return False
        try:
            os.chdir(os.path.join(cwd, self.slug))
            config = Config()
            if config and config.local_config_exists:
                return True
            return False
        finally:
            os.chdir(cwd)




class ExceptionsConfig:
    SUPPRESS_EXCEPTIONS = False


class CnvrgError(Exception):
    def __init__(self, message):
        super(Exception, self).__init__(message)


class CnvrgArgumentsError(CnvrgError):
    def __init__(self, error):
        """
        This error aggregate multiple bad argument errors into one.
        Use when the user gave multiple wrong arguments and should be notified about all of them at once.
        @param error_dict: dict. Key is the argument's name and the value is the message regarding the argument.
        @return:  None
        """
        message = "Bad arguments: "
        if isinstance(error, dict):
            message += "\r\n "
            for key in error.keys():
                if len(error.keys()) > 1:
                    message += ", \r\n"
                message += str(key) + ": " + str(error[key])
        else:
            message += error

        super(CnvrgError, self).__init__(message)


class CnvrgFinalStateReached(CnvrgError):
    def __init__(self, message):
        super(CnvrgError, self).__init__(message)


class CnvrgHttpError(CnvrgError):
    def __init__(self, status_code, message):
        self.status_code = status_code
        super(CnvrgError, self).__init__(message)


class CnvrgFileError(CnvrgError):
    def __init__(self, message):
        super(CnvrgError, self).__init__(message)


class CnvrgRetryError(CnvrgError):
    def __init__(self, message):
        super(CnvrgError, self).__init__(message)


class CnvrgLoginError(CnvrgError):
    def __init__(self, message):
        super(CnvrgError, self).__init__(message)



import fnmatch
import hashlib
import mimetypes
import os
import pathlib
import shutil
from base64 import b64decode
from os import path, walk
from pathlib import Path

import pkg_resources
import requests
from Crypto.Cipher import AES


def unpad(s):
    return s[:-ord(s[len(s) - 1:])]


def filter_by_mimetypes(filters, path):
    """
    mimetypes for files will return in the following format: "TYPE/EXACT_TYPE", for ex. 'image/jpeg'
    This function will receive general types in the filters list (for ex. ['image']), and will only return
    True for files that match any of the filters
    @param filters: Allowed file types.
    @param path: File to test against filters
    @return: True if file passed filters, False if file is filtered
    """
    # This means we have no filters
    if not filters:
        return True

    # This means its a directory / file without extension
    if not mimetypes.guess_type(path)[0]:
        return False

    mime_type = mimetypes.guess_type(path)[0].split('/')[0]
    return mime_type in filters


def decrypt(key, iv, secret):
    cipher = AES.new(key.encode("utf8"), AES.MODE_CBC, b64decode(iv.encode("utf8")))
    secret = b64decode(secret.encode("utf-8"))
    return unpad(cipher.decrypt(secret).decode('utf8'))


def download_file(file_url, file_path):
    """
    Download a file from the provided url and saves it to the requested path
    @param file_url: The URL from which the file will be downloaded
    @param file_path: The target path to which we want to save the file
    @return: None
    """
    sts_file = requests.get(file_url, verify=False)
    create_dir_if_not_exists(file_path)
    with open(file_path, 'wb') as f:
        f.write(sts_file.content)


def create_dir_if_not_exists(local_path):
    """
    Checks if the given file path directory exists and creates it if not.
    @param local_path: File path
    @return: None
    """
    dir_path = os.path.dirname(local_path)
    Path(dir_path).mkdir(parents=True, exist_ok=True)


def get_relative_path(full_path):
    """
    Returns file relative path without '.' suffix
    @param full_path: Path string
    @return: String
    """
    if full_path.startswith("./"):
        full_path = full_path[2:]
    return full_path


def append_trailing_slash(path):
    """
    Appends trailing slash to path - used to differentiate files from directories
    @param path: String representing the directory path
    @return: String
    """
    return path if path.endswith("/") else "{}/".format(path)


def create_cnvrgignore(path):
    """
    Creates .cnvrgignore file in the root project/dataset
    @param path: Root path of the project/dataset
    @return: None
    """
    cnvrg_ignore = pkg_resources.resource_filename('cnvrgv2', 'utils/files/.cnvrgignore')
    shutil.copyfile(cnvrg_ignore, os.path.join(path, '.cnvrgignore'))


def get_cnvrgignore_rules(path):
    """
    Return list of rules after parsing the cnvrgignore file
    @param path: Root path of the project/dataset
    @return: [List] of rules
    """
    cnvrgignore_path = os.path.join(path, '.cnvrgignore')
    rules = ['.cnvrg/', '.tmp/']
    if not os.path.exists(cnvrgignore_path):
        return rules
    with open(cnvrgignore_path, 'r') as f:
        for line in f.readlines():
            if line.startswith("#") or line in ['\n', '\r\n']:
                continue
            rules.append(line.strip())
    return rules


def match_ignore(path, project_root):
    """
    Checks if path match any of the rules
    @param path: file path
    @param project_root: Root path of the project/dataset
    @return: [Bool] true if match rule , false if path didnt match any rules
    """
    rules = get_cnvrgignore_rules(project_root)
    for rule in rules:
        if rule.endswith("/"):
            if pathlib.PurePath(path).match(rule + '*'):
                return True
            elif path.startswith(rule):
                return True
        if pathlib.PurePath(path).match(rule):
            return True
    return False


def cnvrgignore_exists(path):
    """
    Check if .cnvrgignore file exists in project/dataset root path.
    @param path: Root path of the project/dataset
    @return: Bool
    """
    return os.path.exists(os.path.join(path, '.cnvrgignore'))


def get_files_and_dirs_recursive(
        root_dir=".",
        regex="*",
        project_root="",
        filters=None,
        force=False
):
    """
    Recursively traverse a given directory and get all of the relevant files and folders within
    @param root_dir: String representing the directory path
    @param regex: String representing the regex to filter out files/folders
    @param project_root: String representing the project root path
    @param force: Get all files and dirs regardless of .cnvrgignore rules. E.g, used to get .tmp folder
    @return: List
    """
    project_root = append_trailing_slash(project_root) if project_root else ''
    full_paths = []

    for root, dirs, files in walk(root_dir, topdown=False, followlinks=True):
        for name in files:
            file_path = get_relative_path(path.join(root, name)).replace(project_root, "")
            if not force and match_ignore(file_path, project_root):
                continue
            full_paths.append(file_path)
        for name in dirs:
            dir_path = append_trailing_slash(get_relative_path(path.join(root, name))
                                             .replace(project_root, ""))
            if not force and match_ignore(dir_path, project_root):
                continue
            full_paths.append(dir_path)

    # Add the folder itself, if it's an empty one
    if len(os.listdir(root_dir)) == 0:
        dir_path = append_trailing_slash(get_relative_path(root_dir).replace(project_root, ""))
        full_paths.append(dir_path)

    if regex:
        full_paths = fnmatch.filter(full_paths, regex)

    if filters:
        full_paths = [item for item in full_paths if filter_by_mimetypes(filters, item)]

    return full_paths


def chunk_list(list, size):
    """
    Chunks a given list into equally sized chunks (except last chunk)
    @param list: List we want to chunk
    @param size: Integer representing the chunk size
    @return: List of lists
    """
    chunks = []
    for i in range(0, len(list), size):
        chunks.append(list[i:i + size])
    return chunks


def get_file_sha1(full_path):
    """
    Calculates the file SHA1 using its content
    @param full_path: String representing the file path
    @return: String
    """
    BUF_SIZE = 65536
    sha1 = hashlib.sha1()

    with open(full_path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)

    return sha1.hexdigest()


def total_files_size(files):
    total = 0
    for full_path in files:
        if os.path.isfile(full_path):
            total += os.path.getsize(full_path)
    return total



from cnvrgv2.cnvrg import Cnvrg
from cnvrgv2.errors import ExceptionsConfig
from cnvrgv2.modules.project import Project
from cnvrgv2.modules.workflows import Workspace, Webapp, WebappType, Experiment, Endpoint, EndpointKind, EndpointEnvSetup
from cnvrgv2.utils.chart_utils import LineChart, Heatmap, BarChart, ScatterPlot



from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgLoginError
from cnvrgv2.config import Config


def verify_login():
    token, domain, user = Config().get_credential_variables()
    if not (token and domain and user):
        raise CnvrgLoginError(error_messages.CREDENTIALS_MISSING)



import os
import shutil
import tarfile

import requests

from cnvrgv2.cli.utils.progress_bar_utils import init_progress_bar_for_cli
from cnvrgv2.config.routes import LIBRARY_VERSION_BASE, USE_LIBRARY_SUFFIX
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.url_utils import urljoin


class LibraryVersion(DynamicAttributes):
    available_attributes = {
        "version": str,
        "summary": str,
        "author": str,
        "home_page": str,
        "author_email": str,
        "library_id": int,
        "schema_link": str,
        "dependency_link": str,
        "description_link": str,
        "files_download_link": str
    }

    def __init__(self, context=None, version=None, attributes=None):
        # Init data attributes
        super().__init__()

        self._context = Context(context=context)
        self.progress_bar = None

        # Set current context scope to current project
        if version:
            self._context.set_scope(SCOPE.LIBRARY_VERSION, version)

        scope = self._context.get_scope(SCOPE.LIBRARY_VERSION)

        self._proxy = Proxy(context=self._context)
        self._route = LIBRARY_VERSION_BASE.format(scope["organization"], scope["library"], scope["library_version"])
        self._attributes = attributes or {}
        self.version = scope["library_version"]

    def clone(self, progress_bar_enabled=False, working_dir=None):
        """
        Clones the remote library
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        @param working_dir: String the directory that the library will be cloned to
        @return: None
        """
        destination_prefix = './{0}/'.format(working_dir) if working_dir else './'
        destination = "{0}{1}".format(destination_prefix, self._context.library)

        if not os.path.isdir(destination):
            os.mkdir(working_dir)

        response = requests.get(self.files_download_link, allow_redirects=True, stream=True)
        total_length = response.headers.get('content-length')

        # TODO: check download counter goes up every clone

        try:
            self._proxy.call_api(
                route=urljoin(self._route, USE_LIBRARY_SUFFIX),
                http_method=HTTP.PUT,
            )
        except Exception:
            pass  # if libhub failed - I don't care

        tar_path = "{0}.tar.gz".format(self.version)
        file = open(tar_path, 'wb')

        total_length = int(total_length)

        if progress_bar_enabled:
            self.progress_bar = init_progress_bar_for_cli("Downloading", total_length)

        self.progress_bar.start()
        for chunk in response.iter_content(chunk_size=32000):
            delta = len(chunk)
            file.write(chunk)
            if progress_bar_enabled:
                self.progress_bar.throttled_next(delta)
        self.progress_bar.finish()

        file.close()

        tar = tarfile.open(tar_path)
        tar.extractall(destination)
        tar.close()

        if len(os.listdir(destination)) == 1:
            source_dir = os.path.join(destination, os.listdir(destination)[0])
            if os.path.isdir(source_dir):
                file_names = os.listdir(source_dir)

                for file_name in file_names:
                    shutil.move(os.path.join(source_dir, file_name), destination)

                os.removedirs(source_dir)

        if not os.path.isfile(os.path.join(destination, "__init__.py")):
            file = open(os.path.join(destination, "__init__.py"), "w+")
            file.close()

        os.remove(tar_path)



from cnvrgv2.config import routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin


class Ssh(DynamicAttributes):
    available_attributes = {
        "ssh_status": str,
        "pod_name": str,
        "namespace": str,
        "username": str,
        "password": str
    }

    def __init__(self, context, job_id):
        self._context = Context(context=context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)
        self._job_id = job_id
        self._proxy = Proxy(context=self._context)
        self._attributes = {}

        # Split into two routes so that dynamic attributes will work (depends on self._route)
        self._base_route = routes.SSH_BASE.format(scope["organization"], self._job_id)
        self._route = urljoin(self._base_route, routes.SSH_STATUS_SUFFIX)

    def start(self, username, password):
        start_ssh_route = urljoin(self._base_route, routes.SSH_START_SUFFIX)

        attributes = {
            "username": username,
            "password": password
        }

        self._proxy.call_api(
            route=start_ssh_route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="sdk_v2_ssh", attributes=attributes)
        )



import click

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command_without_context
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config import Config
from cnvrgv2.cnvrg import Cnvrg


@click.command()
@click.option(
    '-nc',
    '--no-check-certificate',
    'check',
    flag_value=0,
    help=messages.CONFIG_HELP_CHECK_CERTIFICATE.format("Disable")
)
@click.option(
    '-c',
    '--check-certificate',
    'check',
    flag_value=1,
    help=messages.CONFIG_HELP_CHECK_CERTIFICATE.format("Enable")
)
@click.option('-o', '--organization', help=messages.CONFIG_HELP_ORGANIZTION)
@prepare_command_without_context()
def config(logger, check, organization):
    """
    Change configuration values
    """
    if all(arg is None for arg in [check, organization]):
        click.echo(message=config.get_help(click.get_current_context()))
        logger.info(msg=messages.CONFIG_NO_ARGS_LOG)
        return
    config_file = Config()
    old_organization = config_file.organization
    try:
        if check is not None:
            config_file.check_certificate = bool(check)
        if organization:
            config_file.organization = organization
            config_file.save()
            Cnvrg()  # We want to test the connection with the new organization

        logger.log_and_echo(message=messages.CONFIG_UPDATE_SUCCESS)
    except CnvrgArgumentsError as error:
        config_file.organization = old_organization
        config_file.save()
        logger.log_and_echo(message=str(error), error=True)



from cnvrgv2.config import error_messages, routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.modules.images.utils import ImageLogos
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.json_api_format import JAF


class Image(DynamicAttributes):
    available_attributes = {
        "name": str,
        "tag": str,
        "logo": str,
        "status": str,
        "readme": str,
        "registry_url": str,
        "registry_slug": str,
        "created_by": str,
        "created_at": str,
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current project
        if slug:
            self._context.set_scope(SCOPE.IMAGE, slug)

        scope = self._context.get_scope(SCOPE.IMAGE)

        self.slug = scope["image"]
        self._proxy = Proxy(context=self._context)
        self._route = routes.IMAGE_BASE.format(scope["organization"], self.slug)
        self._attributes = attributes or {}
        self._dockerfile = None

    @property
    def dockerfile(self, refresh=False):
        """
        Returns the image dockerfile
        @param refresh: Force refresh the dockerfile value
        @return: [String] The dockerfile of the image
        """
        if self._dockerfile is None or refresh:
            response = self._proxy.call_api(route="{}/dockerfile".format(self._route), http_method=HTTP.GET)
            self._dockerfile = response.attributes["dockerfile"]

        return self._dockerfile

    def update(self, logo=None, readme=None):
        """
        Updates the current image
        @param logo: [String] The image logo (to be displayed in the UI)
        @param readme: [String] Image readme (will be displayed in the UI)
        @return: Image object
        """

        if not ImageLogos.validate_icon(logo):
            raise CnvrgArgumentsError(error_messages.IMAGE_BAD_LOGO)

        # Take only the attributes we want to update (remove None values)
        attributes = {"logo": logo, "readme": readme}
        attributes = {k: v for k, v in attributes.items() if v is not None}

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type="image", attributes=attributes)
        )

        self._attributes = response.attributes
        return self

    def delete(self):
        """
        Deletes the current image
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)



from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2 import cnvrg
from cnvrgv2.modules.libhub.libraries.library import Library


class LibrariesClient:
    def __init__(self, organization):
        if type(organization) is not cnvrg.Cnvrg:
            raise CnvrgArgumentsError(error_messages.ARGUMENT_BAD_TYPE.format(cnvrg.Cnvrg, type(organization)))
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.LIBRARIES_BASE.format(scope["organization"])

    def list(self, sort="-id"):
        """
        List all projects in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields project objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Library,
            sort=sort,
            pagination_type=None
        )

    def get(self, slug):
        """
        Retrieves a project by the given slug
        @param slug: The slug of the requested project
        @return: Project object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.LIBRARY_GET_FAULTY_SLUG)

        return Library(context=self._context, slug=slug)

    # TODO - need to implement in the future



import click

from prettytable import PrettyTable

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.registries import RegistryTypes


@click.group(name='registry')
def registry_group():
    """
    Manage images in cnvrg
    """
    pass


@registry_group.command()
@click.option('-t', '--title', prompt=messages.REGISTRY_PROMPT_TITLE, help=messages.REGISTRY_HELP_TITLE)
@click.option('-u', '--url', prompt=messages.REGISTRY_PROMPT_URL, help=messages.REGISTRY_HELP_URL)
@click.option('-rt', '--type', default=RegistryTypes.OTHER, help=messages.REGISTRY_HELP_TYPE)
@click.option('-us', '--username', default=None, help=messages.REGISTRY_HELP_USERNAME)
@click.option('-ps', '--password', default=None, help=messages.REGISTRY_HELP_PASSWORD)
@prepare_command()
def create(cnvrg, logger, title, url, type, username, password):
    """
    Creates a new registry
    """

    registry = cnvrg.registries.create(
        url=url,
        type=type,
        title=title,
        username=username,
        password=password
    )

    logger.log_and_echo(messages.REGISTRY_CREATE_SUCCESS.format(registry.title))


@registry_group.command()
@prepare_command()
def list(cnvrg):
    """
    Lists cnvrg registries
    """

    table = PrettyTable()
    table.field_names = ["Slug", "Title", "URL", "Private", "Username", "Type"]
    for field in table.field_names:
        table.align[field] = "l"

    for reg in cnvrg.registries.list():
        table.add_row([reg.slug, reg.title, reg.url, reg.private, reg.username, reg.registry_type])

    click.echo(table)


@registry_group.command()
@click.option('-s', '--slug', help=messages.REGISTRY_HELP_SLUG)
@prepare_command()
def get(cnvrg, logger, slug):
    """
    Retrieves a registry data
    """

    try:
        reg = cnvrg.registries.get(slug=slug)
    except CnvrgError as error:
        logger.log_and_echo(message=str(error), error=True)
        return

    table = PrettyTable()
    table.field_names = ["Slug", "Title", "URL", "Private", "Username", "Type"]
    for field in table.field_names:
        table.align[field] = "l"

    table.add_row([reg.slug, reg.title, reg.url, reg.private, reg.username, reg.registry_type])
    click.echo(table)


@registry_group.command()
@click.option('-s', '--slug', prompt=messages.IMAGE_PROMPT_SLUG, help=messages.IMAGE_HELP_SLUG)
@click.option('-t', '--title', default=None, help=messages.REGISTRY_HELP_TITLE)
@click.option('-u', '--url', default=None, help=messages.REGISTRY_HELP_URL)
@click.option('-us', '--username', default=None, help=messages.REGISTRY_HELP_USERNAME)
@click.option('-ps', '--password', default=None, help=messages.REGISTRY_HELP_PASSWORD)
@prepare_command()
def update(cnvrg, slug, title, url, username, password):
    """
    Updates a registry
    """

    cnvrg.registries.get(slug=slug).update(title=title, url=url, username=username, password=password)


@registry_group.command()
@click.option('-s', '--slug', help=messages.REGISTRY_HELP_SLUG)
@prepare_command()
def delete(cnvrg, logger, slug):
    """
    Deletes a registry
    """

    try:
        reg = cnvrg.registries.get(slug=slug)
    except CnvrgError as error:
        logger.log_and_echo(message=str(error), error=True)
        return

    reg.delete()



import os
import urllib.parse
import yaml
import netrc
from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgError, CnvrgFileError, CnvrgArgumentsError
from pathlib import Path

CONFIG_VERSION = 2
CONFIG_FILE_NAME = "cnvrg.config"
CONFIG_FOLDER_NAME = ".cnvrg"

NETRC_FILE_NAME = ".netrc"
LEGACY_IDX_FILE_NAME = "idx.yml"
LEGACY_CONFIG_FILE_NAME = "config.yml"

LEGACY_NETRC_PATH = os.path.join(os.path.expanduser("~"), NETRC_FILE_NAME)  # TODO: might break on windows
GLOBAL_CNVRG_PATH = os.path.join(os.path.expanduser("~"), CONFIG_FOLDER_NAME)  # TODO: might break on windows


class Config:

    global_config_attributes = {
        "organization": str,
        "domain": str,
        "token": str,
        "user": str,
        "check_certificate": bool,
        "keep_duration_days": int,
        "version": float
    }

    local_config_attributes = {
        "git": bool,
        "commit_sha1": str,
        "project_slug": str,
        "library_slug": str,
        "dataset_slug": str,
    }

    def __init__(self):
        self.is_capi = None
        self.local_config = {}
        self.global_config = {}
        if self.local_cnvrg_path:
            self.root = str(Path(self.local_cnvrg_path).parent)
        else:
            self.root = ""

        # Load all relevant config config files
        self._load_legacy_config()
        self._load_global_config()
        self._load_local_config()

    @property
    def data_owner_slug(self):
        return self.project_slug or self.dataset_slug

    @property
    def global_cnvrg_path(self):
        return GLOBAL_CNVRG_PATH

    @property
    def local_cnvrg_path(self):
        path = os.getcwd()
        # stop if we reach the home path to avoid the global .cnvrg
        while str(path) not in ['~', '/', str(Path.home())]:
            joined_path = os.path.join(str(path), CONFIG_FOLDER_NAME)
            if os.path.exists(joined_path):
                return joined_path
            path = Path(path).parent
        return ""

    @property
    def local_config_file_path(self):
        return os.path.join(self.local_cnvrg_path, CONFIG_FILE_NAME)

    @property
    def global_config_file_path(self):
        return os.path.join(self.global_cnvrg_path, CONFIG_FILE_NAME)

    @property
    def legacy_global_config_file_path(self):
        return os.path.join(self.global_cnvrg_path, LEGACY_CONFIG_FILE_NAME)

    @property
    def legacy_local_config_file_path(self):
        return os.path.join(self.local_cnvrg_path, LEGACY_CONFIG_FILE_NAME)

    @property
    def local_config_exists(self):
        return os.path.exists(self.local_config_file_path)

    @property
    def global_config_exists(self):
        return os.path.exists(self.global_config_file_path)

    def _load_global_config(self):
        if self.global_config_exists:
            global_config = yaml.safe_load(open(self.global_config_file_path, "r")) or {}
            if global_config.get("token"):
                self.is_capi = True

            for key in Config.global_config_attributes.keys():
                default = False if Config.global_config_attributes[key] == bool else None
                self.global_config[key] = global_config.get(key, default) or self.global_config.get(key, default)

    def _load_local_config(self):
        if self.local_config_exists:
            local_config = yaml.safe_load(open(self.local_config_file_path, "r")) or {}
            for key in Config.local_config_attributes.keys():
                default = False if Config.local_config_attributes[key] == bool else None
                self.local_config[key] = local_config.get(key, default) or self.local_config.get(key, default)

    def _load_legacy_config(self):
        legacy_idx_file_path = os.path.join(self.local_cnvrg_path, LEGACY_IDX_FILE_NAME)

        legacy_global_config_exists = os.path.exists(self.legacy_global_config_file_path)
        legacy_netrc_exists = os.path.exists(LEGACY_NETRC_PATH)

        legacy_idx_exists = os.path.exists(legacy_idx_file_path)
        legacy_config_exists = os.path.exists(self.legacy_local_config_file_path)

        organization = None
        # We want to try and load legacy config file BEFORE the new config file so we can override if needed
        if legacy_global_config_exists:
            legacy_global_config = yaml.safe_load(open(self.legacy_global_config_file_path, "r")) or {}
            organization = legacy_global_config.get(":owner", None) or legacy_global_config.get("owner", None)
            api = legacy_global_config.get(":api", None) or legacy_global_config.get("api", None)
            uri = urllib.parse.urlparse(api)
            self.domain = '{uri.scheme}://{uri.netloc}/'.format(uri=uri)

        if legacy_netrc_exists:
            netrc_file = netrc.netrc()
            auth_tokens = netrc_file.authenticators("cnvrg.io")
            if auth_tokens:
                # 0 is login, 1 is account, 2 is password
                self.user = auth_tokens[0]
                self.token = auth_tokens[2]
                self.is_capi = False

        if legacy_config_exists:
            legacy_config = yaml.safe_load(open(self.legacy_local_config_file_path, "r")) or {}
            organization = legacy_config.get(":owner", None) or legacy_config.get("owner", None) or organization
            self.project_slug = legacy_config.get("project_slug", None) or legacy_config.get(":project_slug", None)
            self.library_slug = legacy_config.get("project_slug", None) or legacy_config.get(":library_slug", None)
            self.dataset_slug = legacy_config.get("dataset_slug", None) or legacy_config.get(":dataset_slug", None)
            self.git = legacy_config.get(":git", False) or legacy_config.get("git", False)

        if legacy_idx_exists:
            legacy_idx_config = yaml.safe_load(open(legacy_idx_file_path, "r")) or {}
            self.commit_sha1 = legacy_idx_config.get(":commit", None) or legacy_idx_config.get("commit", None)

        self.organization = organization

    def __getattr__(self, name):
        """
        Returns an attribute value from the local/global config attributes
        @param name: Name of the attribute to return
        @return: The value of the given attribute
        """
        is_global = name in Config.global_config_attributes.keys()
        is_local = name in Config.local_config_attributes.keys()

        if not (is_global or is_local):
            raise AttributeError("type object {} has no attribute {}".format(self.__class__.__name__, name))

        if is_global:
            return self.global_config.get(name, None)
        else:
            return self.local_config.get(name, None)

    def __setattr__(self, name, value):
        """
        Sets an attribute to global/local config
        @param name: Attribute name
        @param value: Attribute value
        @return: None
        """
        is_global = name in Config.global_config_attributes.keys()
        is_local = name in Config.local_config_attributes.keys()

        config_available_attributes = Config.global_config_attributes if is_global else Config.local_config_attributes

        if not (is_global or is_local):
            super().__setattr__(name, value)
        else:
            value_type = type(value)

            expected_type = config_available_attributes[name]
            if value_type != expected_type and value is not None:
                bad_format_message = error_messages.ARGUMENT_BAD_TYPE.format(expected_type, value_type)
                raise CnvrgArgumentsError({name: bad_format_message})

            if is_global:
                self.global_config[name] = value
            else:
                self.local_config[name] = value

    def update(self, local_config_path=None, **kwargs):
        """
        Updates the config file and changes the attributes of the current object.
        Unknown keys will be ignored.
        @param local_config_path: Override the local config path - can be used to create new cnvrg project or dataset
        @param kwargs: Dict of new values
        @return: None
        """
        updated_global_attributes = {}
        updated_local_attributes = {}

        for key, value in kwargs.items():
            if key in Config.global_config_attributes.keys():
                updated_global_attributes[key] = value
            if key in Config.local_config_attributes.keys():
                updated_local_attributes[key] = value

        self.global_config = {
            **self.global_config,
            **updated_global_attributes
        }

        self.local_config = {
            **self.local_config,
            **updated_local_attributes
        }

        self.save(local_config_path=local_config_path)

    def save(self, local_config_path=None):
        """
        Updates the config file with the current attributes
        @return: None
        """

        if local_config_path is None:
            if self.local_cnvrg_path:
                local_config_path = self.local_cnvrg_path
            else:
                local_config_path = os.path.join(os.getcwd(), CONFIG_FOLDER_NAME)

        path_to_attrs = {}
        if self.global_config:
            path_to_attrs[os.path.join(GLOBAL_CNVRG_PATH, CONFIG_FILE_NAME)] = self.global_config
        if self.local_config:
            path_to_attrs[os.path.join(local_config_path, CONFIG_FILE_NAME)] = self.local_config

        for config_path, new_config in path_to_attrs.items():
            if not os.path.exists(os.path.dirname(config_path)):
                os.makedirs(os.path.dirname(config_path))

            if not os.path.exists(config_path):
                open(config_path, "w").close()

            with open(config_path, 'r+') as config_file:
                config = yaml.safe_load(config_file) or {}
                config = {**config, **new_config}
                config_file.seek(0)
                config_file.truncate(0)
                yaml.dump(config, config_file)

    def remove_config_fields(self, *args):
        """
        Removes fields from the config file according to given keys
        @param args: keys to remove
        @return: None
        """

        paths_to_keys = {}
        for key in args:
            key_exists = False
            if key in Config.global_config_attributes.keys():
                paths_to_keys.setdefault(self.global_config_file_path, []).append(key)
                key_exists = True
            if key in Config.local_config_attributes.keys():
                paths_to_keys.setdefault(self.local_config_file_path, []).append(key)
                key_exists = True

            if not key_exists:
                raise CnvrgError(error_messages.FAULTY_KEY.format(key))

        for config_path, keys_to_remove in paths_to_keys.items():
            if not os.path.exists(os.path.dirname(config_path)):
                raise CnvrgFileError(error_messages.CONFIG_FILE_MISSING)

            with open(config_path, 'r+') as config_file:
                config = yaml.safe_load(config_file) or {}
                for key in keys_to_remove:
                    config.pop(key, None)
                config_file.seek(0)
                config_file.truncate(0)
                yaml.dump(config, config_file)

    def get_credential_variables(self):
        """
        Returns the variables forming user credentials
        @return: token, domain, user
        """
        if not self.global_config:
            return False, False, False

        # Credential variables
        token = self.token
        domain = self.domain
        user = self.user

        return token, domain, user



from cnvrgv2.config import routes
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes


class Query(DynamicAttributes):
    available_attributes = {
        "slug": str,
        "name": str,
        "query": str,
        "commit_sha1": str,
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current query
        if slug:
            self._context.set_scope(SCOPE.QUERY, slug)

        scope = self._context.get_scope(SCOPE.QUERY)

        self._proxy = Proxy(context=self._context)
        self._route = routes.QUERY_BASE.format(scope["organization"], scope["dataset"], scope["query"])
        self._attributes = attributes or {}
        self.slug = scope["query"]

    def delete(self):
        """
        Deletes the current query
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)



from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.queries.query import Query
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator


class QueriesClient:
    def __init__(self, context=None):
        self._context = Context(context=context)
        scope = self._context.get_scope(SCOPE.DATASET)

        self._proxy = Proxy(context=self._context)
        self._route = routes.QUERIES_BASE.format(scope["organization"], scope["dataset"])

    def list(self, sort="-id"):
        """
        List all queries in a specific query
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields query objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Query,
            sort=sort
        )

    def get(self, slug):
        """
        Retrieves a query by the given slug
        @param slug: The slug of the requested dataset
        @return: Query object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.QUERY_GET_FAULTY_SLUG)

        return Query(context=self._context, slug=slug)

    def create(self, name, query, commit_sha1=None):
        """
        Creates a new query with the given name
        @param name: Name of the new query
        @param query: Query string of the new query
        @param commit_sha1: Sha1 of the query commit
        @return: The newly created query object
        """
        if not name or not isinstance(name, str):
            raise CnvrgArgumentsError(error_messages.QUERY_CREATE_FAULTY_NAME)

        if not query or not isinstance(query, str):
            raise CnvrgArgumentsError(error_messages.QUERY_CREATE_FAULTY_QUERY)

        attributes = {
            "name": name,
            "query_raw": query,
            "commit_sha1": commit_sha1
        }
        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="query", attributes=attributes)
        )

        slug = response.attributes['slug']
        return Query(context=self._context, slug=slug, attributes=response.attributes)



# VERSION
VERSION = "v2/version/"

# ORGANIZATION
ORGANIZATION_BASE = "v2/{0}/"
ORGANIZATION_SETTINGS = "v2/{0}/settings"
ORGANIZATION_USERS = "v2/{0}/users"
ORGANIZATION_MEMBERS = "v2/{0}/members"
ORGANIZATION_CREATE = "v2/organizations"

# USER
USER_BASE = "v2/users/"
USER_LOGIN = "v2/users/sign_in"
USER_CURRENT = "v2/users/me"
USER_MEMBERSHIPS = "v2/users/me/memberships/"

# PROJECT
PROJECTS_BASE = "v2/{0}/projects/"
PROJECT_BASE = "v2/{0}/projects/{1}"
PROJECT_CLONE_SUFFIX = "clone"
PROJECT_BULK_DELETE_SUFFIX = "delete_projects"
PROJECT_DELETE_SECRET = "delete_secrets"
PROJECT_UPDATE_SECRET = "update_secrets"

# DATASET
DATASETS_BASE = "v2/{0}/datasets/"
DATASET_BASE = "v2/{0}/datasets/{1}"

# LIBHUB
LIBHUB_BASE = "v2/libhub/"
LIBRARIES_BASE = "v2/{0}/libhub/libraries/"
LIBRARY_BASE = "v2/{0}/libhub/libraries/{1}/"
USE_LIBRARY_SUFFIX = "use"
LIBRARY_VERSIONS_BASE = "v2/{0}/libhub/libraries/{1}/versions/"
LIBRARY_VERSION_BASE = "v2/{0}/libhub/libraries/{1}/versions/{2}"

# COMMIT
DATASET_COMMITS_BASE = "v2/{0}/datasets/{1}/commits/"
DATASET_COMMIT_BASE = "v2/{0}/datasets/{1}/commits/{2}"
PROJECT_COMMITS_BASE = "v2/{0}/projects/{1}/commits/"
PROJECT_COMMIT_BASE = "v2/{0}/projects/{1}/commits/"
COMMIT_REMOVE_FILES = "commits/{0}/remove_files?page[size]=1000&sort=id"

# QUERIES
QUERY_BASE = "v2/{0}/datasets/{1}/queries/{2}"
QUERIES_BASE = "v2/{0}/datasets/{1}/queries"

# IMAGE
IMAGES_BASE = "v2/{0}/images/"
IMAGE_BASE = "v2/{0}/images/{1}"

# REGISTRY
REGISTRIES_BASE = "v2/{0}/registries/"
REGISTRY_BASE = "v2/{0}/registries/{1}"

# VOLUME
VOLUMES_BASE = "v2/{0}/projects/{1}/volumes"
VOLUME_BASE = "v2/{0}/projects/{1}/volumes/{2}"

# FLOW
FLOWS_BASE = "v2/{0}/projects/{1}/flows/"
FLOW_BASE = "v2/{0}/projects/{1}/flows/{2}/"
FLOW_SET_SCHEDULE = "start-cron"
FLOW_STOP_SCHEDULE = "stop-cron"
CREATE_WITH_YAML = "create-with-yaml"
RUN_FLOW = "run"
FLOW_TOGGLE_WEBHOOK = "toggle-webhook"
FLOW_TOGGLE_DATASET_UPDATE = "toggle-dataset-update"
FLOW_LATEST_END_COMMITS = "latest-end-commits"
FLOW_REVIEWERS = "reviewers"

# FLOW_VERSION
FLOW_VERSIONS_BASE = "v2/{0}/projects/{1}/flows/{2}/flow-versions"
FLOW_VERSION_BASE = "v2/{0}/projects/{1}/flows/{2}/flow-versions/{3}"
FLOW_VERSION_INFO = "info"
FLOW_VERSION_STOP = "stop"

# WORKFLOW
WORKFLOW_STANDALONE_BASE = "v2/{0}/workflows/"
WORKFLOWS_BASE = "v2/{0}/projects/{1}/workflows/"
WORKFLOWS_BASE_NEW = "v2/{0}/projects/{1}/{2}/"
WORKFLOW_BASE = "v2/{0}/projects/{1}/workflows/{2}/"
WORKFLOW_STOP_SUFFIX = "stop"
WORKFLOW_RESTART_SUFFIX = "restart"
WORKFLOW_SYNC_SUFFIX = "sync"
WORKFLOW_LIST_TYPE_SUFFIX = "?filter={{" \
                            "\"operator\":\"AND\"," \
                            "\"conditions\":[{{\"key\":\"job_type\",\"operator\":\"is\",\"value\":\"{0}\"}}]" \
                            "}}"
WORKflOW_START_SUFFIX = "start"
WORKflOW_START_TENSORBOARD_SUFFIX = "start_tensorboard"
WORKflOW_STOP_TENSORBOARD_SUFFIX = "stop_tensorboard"
WORKFLOW_WRITE_LOGS_SUFFIX = "write_logs"
WORKFLOW_GET_LOGS_SUFFIX = "logs"
WORKFLOW_TAG_SUFFIX = "tags"
WORKFLOW_UPDATE_SUFFIX = "update"
WORKFLOW_ARTIFACTS_SUFFIX = "artifacts"

# WORKSPACE
WORKSPACES_BASE = "v2/{0}/projects/{1}/workspaces/"
WORKSPACE_BASE = "v2/{0}/projects/{1}/workspaces/{2}/"
WORKSPACE_BUILD_IMAGE_SUFFIX = "build_image"
WORKSPACE_ACTIVATE_BUILD_IMAGE_ON_STOP_SUFFIX = "activate_build_image_on_stop"
WORKSPACE_DISABLE_BUILD_IMAGE_ON_STOP_SUFFIX = "disable_build_image_on_stop"

# EXPERIMENT
EXPERIMENTS_BASE = "v2/{0}/projects/{1}/experiments/"
EXPERIMENT_BASE = "v2/{0}/projects/{1}/experiments/{2}/"
EXPERIMENT_RERUN_SUFFIX = "rerun"
EXPERIMENT_FINISH_SUFFIX = "finish"
EXPERIMENT_CHARTS_SUFFIX = "charts"
EXPERIMENT_GET_UTILIZATION_SUFFIX = "get_utilization"
EXPERIMENT_CHART_SUFFIX = "charts/{0}"
EXPERIMENT_WRITE_LOGS = "write_logs"
EXPERIMENT_LOGS_SUFFIX = "logs?offset=-1&search=&filter=[\"all\"]"

# WEBAPP
WEBAPPS_BASE = "v2/{0}/projects/{1}/webapps/"
WEBAPP_BASE = "v2/{0}/projects/{1}/webapps/{2}/"

# ENDPOINT
ENDPOINTS_BASE = "v2/{0}/projects/{1}/endpoints"
ENDPOINT_BASE = "v2/{0}/projects/{1}/endpoints/{2}/"
UPDATE_MODEL_VERSION = "update_model"
ENDPOINT_FEEDBACK_LOOP = "feedback_loop"
ROLLBACK_MODEL_VERSION = "rollback_session/{0}"
ENDPOINT_SAMPLE_CODE = "sample_code"
ENDPOINT_REPLICAS = "replicas"
ENDPOINT_POLL_CHARTS = "poll_charts"
ENDPOINT_STATUS = "status"
ENDPOINT_BATCH_SCALE_UP = "batch_scale_up"
ENDPOINT_BATCH_SCALE_DOWN = "batch_scale_down"
ENDPOINT_RULES = "rules"

# COMMON
GET_BY_NAME_SUFFIX = "get_by_name"

# RESOURCES
MACHINES_BASE = "v2/{0}/resources/machines/computes"
SPARK_DRIVERS_BASE = "v2/{0}/resources/spark_driver"
CLUSTERS_BASE = "v2/{0}/resources"
MACHINE_BASE = "v2/{0}/resources/machines/computes/{1}"
SPARK_DRIVER_BASE = "v2/{0}/resources/spark_driver/{1}"
CLUSTER_BASE = "v2/{0}/resources/{1}"

# TEMPLATE
TEMPLATES_BASE = "v2/{0}/resources/{1}/{2}/computes"
TEMPLATE_BASE = "v2/{0}/resources/{1}/{2}/computes/{3}"

# SSH
SSH_BASE = "v2/{0}/job_ssh/{1}"
SSH_START_SUFFIX = "start"
SSH_STATUS_SUFFIX = "status"

# GIT
GIT = "git"



from cnvrgv2.config import routes, error_messages
from cnvrgv2.config.error_messages import FAULTY_KEY
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.data_owner import DataOwner
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin
from cnvrgv2.utils.validators import attributes_validator, validate_secret_key
from cnvrgv2.errors import CnvrgArgumentsError, CnvrgFileError


class ProjectSettings(DataOwner):
    available_attributes = {
        "title": str,
        "default_image": str,
        "privacy": str,
        "mount_folders": list,
        "env_variables": list,
        "check_stuckiness": bool,
        "max_restarts": int,
        "stuck_time": int,
        "autosync": bool,
        "sync_time": int,
        "default_computes": list,
        "use_org_deploy_key": bool,
        "deploy_key": str,
        "webhooks_url": str,
        "slack_webhooks": bool,
        "slack_webhook_channel": str,
        "command_to_execute": str,
        "run_tensorboard_by_default": bool,
        "run_jupyter_by_default": bool,
        "email_on_error": bool,
        "email_on_success": bool,
        "working_directory": str,
        "requirements_path": str,
        "secrets": list,
        "project_idle_time": float,
        "is_git": bool,
        "git_repo": str,
        "git_branch": str,
        "private_repo": bool,
        "description": str,
        "tags": list,
        "collaborators": list,
        "git_access_token": bool,
        "output_dir": str
    }

    def __init__(self, project):
        self._context = Context(context=project._context)
        scope = self._context.get_scope(SCOPE.PROJECT)

        self._proxy = Proxy(context=self._context)

        org_base = routes.PROJECT_BASE.format(scope["organization"], scope["project"])
        self._route = urljoin(org_base, "settings")

        self._attributes = {}

    def save(self):
        """
        Save the local settings in the current project
        @return: None
        """
        valid_attributes = {**self._attributes}

        # Secrets are being updates in a separate function
        if valid_attributes.get("secrets") is not None:
            del valid_attributes["secrets"]

        self.update(**valid_attributes)

    def update(self, **kwargs):
        """
        Updates current project's settings with the given params
        @param kwargs: any param out of the available attributes can be sent
        @return: None
        """
        attributes_validator(
            available_attributes=ProjectSettings.available_attributes,
            attributes=kwargs,
        )

        if "secrets" in kwargs:
            raise CnvrgArgumentsError(error_messages.SECRETS_NOT_SUPPORTED.format("secrets"))

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="settings", attributes={**self._attributes, **kwargs})
        )

        response_attributes = response.attributes
        response_attributes.pop('slug', None)
        self._attributes = response_attributes

    def update_secret(self, secret_key, secret_value, force=False):
        """
        update or add a single secret
        @param secret_key: [String] Secret Key
        @param secret_value: [String] Secret Value
        @param force: Override existing secrets
        @return: None
        """
        if not validate_secret_key(secret_key):
            raise CnvrgArgumentsError(FAULTY_KEY.format(secret_key))

        update_secret_url = urljoin(self._route, routes.PROJECT_UPDATE_SECRET)

        self._proxy.call_api(
            route=update_secret_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="settings", attributes={
                "secret_key": secret_key,
                "secret_value": secret_value,
                "force": force
            })
        )

    def delete_secret(self, secrets):
        """
        delete single or list of secrets
        @param secrets: [List] of Keys
        @return: None
        """
        if isinstance(secrets, str):
            secrets = [secrets]

        delete_secret_url = urljoin(self._route, routes.PROJECT_DELETE_SECRET)

        self._proxy.call_api(
            route=delete_secret_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="settings", attributes={"secrets": secrets})
        )

    def setup_git(self, git_repo, branch, token=None, ssh_key=None, ssh_key_type="plain"):
        if all([token, ssh_key]):
            raise CnvrgArgumentsError(error_messages.CANT_USE_BOTH_SSH_AND_TOKEN)

        if ssh_key_type not in ("plain", "file"):
            raise CnvrgArgumentsError(error_messages.INVALID_SSH_KEY_TYPE)

        if ssh_key_type == "file" and ssh_key is not None:
            try:
                with open(ssh_key, "r") as file:
                    ssh_key = file.read()
            except Exception:
                raise CnvrgFileError(error_messages.CANT_READ_SSH_KEY_FILE)

        if ssh_key is not None:
            authentication_method = "sshKey"
        elif token is not None:
            authentication_method = "token"
        else:
            authentication_method = "public"

        self._proxy.call_api(route=urljoin(self._route, routes.GIT), http_method=HTTP.POST, payload={
            "git_repo": git_repo,
            "branch": branch,
            "token": token,
            "ssh_key": ssh_key,
            "authentication_method": authentication_method
        })
        self.reload()

    def remove_git(self):
        self._proxy.call_api(route=urljoin(self._route, routes.GIT), http_method=HTTP.DELETE)
        self._attributes.pop("git_repo", None)
        self._attributes.pop("git_branch", None)
        self._attributes.pop("is_git", None)



from cnvrgv2.data.data_loader import DataLoader
from cnvrgv2.data.file_compare import FileCompare
from cnvrgv2.data.file_uploader import FileUploader
from cnvrgv2.data.file_downloader import FileDownloader
from cnvrgv2.data.local_file_deleter import LocalFileDeleter
from cnvrgv2.data.remote_file_deleter import RemoteFileDeleter
from cnvrgv2.data.artifacts_downloader import ArtifactsDownloader



import json
import os
from pathlib import Path

import cnvrgv2.utils.storage_utils as storage
from cnvrgv2.data.remote_files_handler import RemoteFilesHandler
from cnvrgv2.proxy import HTTP
from cnvrgv2.utils.url_utils import urljoin


class LocalFileDeleter(RemoteFilesHandler):
    def __init__(self, data_owner, num_workers=40, queue_size=5000, chunk_size=1000, force=False, mode="files"):
        """
        Multithreaded local file deleter - deleting files from local if they deleted on server
        @param data_owner: Cnvrg dataset / project object
        @param num_workers: Number of threads to handle files
        @param queue_size: Max number of file meta to put in queue
        @param chunk_size: File meta chunk size to fetch from the server
        @param force: Force rewrite existing files
        """
        self.mode = mode
        super().__init__(
            data_owner,
            num_workers=num_workers,
            queue_size=queue_size,
            chunk_size=chunk_size,
            force=force,
            base_commit_sha1=data_owner.last_commit,
            commit_sha1=data_owner.local_commit
        )

    def _collector_function(self, page_after=None):
        """
        Function to collect files that should be delete locally
        @param page_after: The id of the next file that the iteration of the pagination should start from
        @return: Should return array of files metadata
        """
        if not self.base_commit_sha1:
            raise AttributeError("base commit must be sent")

        data = {
            "base_commit_sha1": self.base_commit_sha1,
            "get_deleted": True,
            "mode": self.mode,
            "filter": json.dumps({
                "operator": 'OR',
                "conditions": [
                    {
                        "key": 'fullpath',
                        "operator": 'like',
                        "value": "*",
                    }
                ],
            })
        }

        response = self.data_owner._proxy.call_api(
            route="{}?{}".format(
                urljoin(self.data_owner._route, "commits", self.commit_sha1, "compare"),
                "page[after]={}&page[size]=1000&sort=id".format(page_after)
            ),
            http_method=HTTP.GET,
            payload=data
        )

        file_dict = []
        for file in response.items:
            file_dict.append(dict(file.attributes))

        return {
            "file_dict": file_dict,
            "total_files": response.meta["total"],
            "next": response.meta["next"]
        }

    def _handle_file_function(self, local_path, **kwargs):
        """
        Function that creates fullpath.deleted file in .tmp that in the end of the
        whole task will cause deletion of fullpath
        @param local_path: File location locally
        @param kwargs: Needs to be fullpath of the file
        @return: None
        """

        tmp_folder_name = "{}/.tmp".format(self.data_owner.working_dir)

        if self.mode == "folders":
            deleted_folder_path = "{}/{}".format(tmp_folder_name, kwargs["fullpath"].rstrip("/"))
            # Check if the folder that should be deleted, is already created in tmp
            # (e.g. through marking a file inside it as deleted)
            if os.path.isdir(deleted_folder_path):
                os.rename(deleted_folder_path, deleted_folder_path + ".deleted")
            else:
                Path(deleted_folder_path + ".deleted").mkdir(parents=True, exist_ok=True)

        else:
            storage.create_dir_if_not_exists("{}/{}".format(tmp_folder_name, kwargs["fullpath"]))
            new_path = "{}/{}.deleted".format(tmp_folder_name, kwargs["fullpath"])
            open(new_path, 'a').close()

        self.handle_queue.task_done()
        self.progress_queue.put(local_path)



import re
import os
from cnvrgv2.config import error_messages
from cnvrgv2.modules.users.user import ROLES
from cnvrgv2.errors import CnvrgArgumentsError


def validate_url(url):
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    return re.match(regex, url) is not None


def validate_email(email):
    regex = r'^[a-zA-Z0-9]+[\._]?[a-zA-Z0-9]+[@]\w+[.]\w{2,5}$'
    return re.search(regex, email) is not None


def validate_user_role(role):
    return role in ROLES.ALL_ROLES


def validate_secret_key(key):
    regex = r'^[a-zA-Z_][a-zA-Z0-9_]*$'
    return re.match(regex, key) is not None


def validate_username(username):
    regex = r'^[A-Za-z0-9]+$'
    return re.search(regex, username) is not None


def validate_directory_name(paths):
    regex = r"^[a-zA-Z0-9_.()'!\/-]+$"
    for path in paths:
        is_dir = os.path.isdir(path)
        if is_dir and not re.match(regex, path):
            raise CnvrgArgumentsError(error_messages.NOT_A_VALID_DIRECTORY_NAME.format(path))


def validate_types_in_list(input_list, input_type):
    """
    Validates that the input is a list of objects from the given type
    @param input_list: The input (should be a list of input_type objects)
    @param input_type: The type of the objects in input_list
    @return: Whenever input_list is a list of input_types
    """
    if not isinstance(input_list, list):
        return False
    for obj in input_list:
        if not isinstance(obj, input_type):
            return False
    return True


def attributes_validator(
        available_attributes,
        attributes,
        required_values=None,
):
    """
    @param available_attributes: Attributes available to the object that is being validated
    @param attributes: Attributes to validate
    @param required_values: Attributes that must be presented in the action
    @return: void, will raise errors on failed validations
    """

    required_values = required_values or []

    # Validates all the required attributes were received (and are not None or equivalent)
    for key in required_values:
        if key not in attributes.keys() or not attributes[key]:
            raise CnvrgArgumentsError(error_messages.MISSING_REQUIRED_VALUE.format(key))

    for key, value in attributes.items():
        # Validates that only allowed attributes were received
        if key == 'slug' or key not in available_attributes.keys():
            raise CnvrgArgumentsError(error_messages.FAULTY_KEY.format(key))
        # Validate received attribute types
        if value and not isinstance(value, available_attributes[key]):
            raise CnvrgArgumentsError(error_messages.FAULTY_VALUE.format(key))


# COMPUTES
def validate_gpu(gpu_obj):
    AVAILABLE_MIG_TYPES = [
        None,
        'mig-1g.5g',
        'mig-2g.20g',
        'mig-3g.25g'
    ]

    gpu = gpu_obj["count"]
    mig_device = gpu_obj["mig_device"]
    if mig_device not in AVAILABLE_MIG_TYPES:
        raise CnvrgArgumentsError(error_messages.TEMPLATE_FAULTY_MIG_VALUE)
    if gpu and not isinstance(gpu, float):
        raise CnvrgArgumentsError(error_messages.TEMPLATE_FAULTY_GPU.format(gpu))
    if mig_device and not isinstance(mig_device, str):
        raise CnvrgArgumentsError(error_messages.TEMPLATE_FAULTY_MIG.format(mig_device))


def validate_template_type(template_type):
    AVAILABLE_TEMPLATE_TYPES = [
        'regular',
        'mpi',
        'kubernetes',
        'pytorch',
        'spark_on_kubernetes'
    ]

    if template_type and template_type not in AVAILABLE_TEMPLATE_TYPES:
        raise CnvrgArgumentsError(error_messages.TEMPLATE_FAULTY_TYPE)



import atexit
import threading
import multiprocessing


class AsyncHandler:
    _open_threads = []
    _open_processes = []
    _registered = False

    @classmethod
    def check_registered(cls):
        if not cls._registered:
            atexit.register(AsyncHandler.wait_fot_threads)
            cls._registered = True

    @classmethod
    def register_async(cls, obj, mode):
        cls.check_registered()
        if mode == "thread":
            cls._open_processes.append(obj)
        elif mode == "process":
            cls._open_threads.append(obj)

    @classmethod
    def wait_fot_threads(cls):
        for t in cls._open_threads:
            t.join()

        for p in cls._open_processes:
            p.join()


def run_async(mode="thread"):
    """
    This decorator with arguments can be used to run functions asynchronously
    :param mode: thread/process
    return: decorator function
    """

    # Validate developer errors
    assert mode in ["thread", "process"], "run_async: mode must be either thread or process"

    # Build the decorator function
    def decorator(f):
        def wrapping_function(*args, **kwargs):
            if mode == "thread":
                obj = threading.Thread(target=f, args=args, kwargs=kwargs)
            else:
                obj = multiprocessing.Process(target=f, args=args, kwargs=kwargs)

            obj.start()
            AsyncHandler.register_async(obj, mode)
            return obj

        return wrapping_function
    return decorator



import os
import json
import time

from queue import Empty

from cnvrgv2.proxy import HTTP
from cnvrgv2.utils.url_utils import urljoin
from cnvrgv2.data.remote_files_handler import RemoteFilesHandler


class RemoteFileDeleter(RemoteFilesHandler):
    def __init__(self, data_owner, commit, queue_size=5000, chunk_size=1000, force=False, files=None):
        """
        Multithreaded remote file deleter - deleting files from server if they deleted locally
        @param data_owner: Cnvrg dataset / project object
        @param chunk_size: File meta chunk size to fetch from the server
        @param force: Force rewrite existing files
        @param queue_size: Max number of file meta to put in queue
        """
        self.files = files
        self.mode = "files"
        self.commit = commit
        self.files_to_delete = []
        self.folders_to_delete = []

        total_files = data_owner._proxy.call_api(
            route="{}?{}".format(
                urljoin(data_owner._route, "commits", data_owner.local_commit, "files"),
                "page[after]={}&page[size]={}&sort=id&mode={}".format('', 0, "files")
            ),
            http_method=HTTP.GET
        ).meta["total"]

        total_folders = data_owner._proxy.call_api(
            route="{}?{}".format(
                urljoin(data_owner._route, "commits", data_owner.local_commit, "files"),
                "page[after]={}&page[size]={}&sort=id&mode={}".format('', 0, "folders")
            ),
            http_method=HTTP.GET
        ).meta["total"]

        super().__init__(
            data_owner,
            num_workers=1,
            queue_size=queue_size,
            chunk_size=chunk_size,
            force=force
        )

        self.total_files = total_files + total_folders

    def _collector_function(self, page_after=None, folders=None):
        """
        Function to collect exists files metadata from server to check if they deleted locally
        @param page_after: The id of the next file that the iteration of the pagination should start from
        @return: Should return array of files metadata
        """
        # Reset the cursor
        page_after = 0 if page_after == -100 else page_after

        if self.files:
            files_index = page_after - 1
            return {
                "file_dict": self.files[files_index:(files_index + self.chunk_size + 1)],
                "total_files": len(self.files),
                "next": page_after + self.chunk_size
            }
        else:
            data = {
                "filter": json.dumps({
                    "operator": 'OR',
                    "conditions": [
                        {
                            "key": 'fullpath',
                            "operator": 'like',
                            "value": "*",
                        }
                    ],
                })
            }

            response = self.data_owner._proxy.call_api(
                route="{}?{}".format(
                    urljoin(self.data_owner._route, "commits", self.data_owner.local_commit, "files"),
                    "page[after]={}&page[size]={}&sort=id&mode={}".format(page_after, self.chunk_size, self.mode)
                ),
                http_method=HTTP.GET,
                payload=data
            )

            # Once we are out of files we want to move to deleting folders
            if response.meta["next"] == '' and self.mode == "files":
                self.mode = "folders"
                return {
                    "file_dict": [],
                    "total_files": 0,
                    "next": -100  # placeholder in order to nullify the cursor
                }

            file_dict = []
            for file in response.items:
                file_attr = dict(file.attributes)
                fullpath = file_attr["fullpath"]
                file_attr["type"] = self.mode
                local_path = "{}/{}".format(self.data_owner.working_dir, fullpath)
                if os.path.isfile(local_path) and self.mode == "files":
                    self.progress_queue.put(file_attr)
                elif os.path.isdir(local_path) and self.mode == "folders":
                    self.progress_queue.put(file_attr)
                else:
                    file_dict.append(file_attr)

            return {
                "file_dict": file_dict,
                "total_files": response.meta["total"],
                "next": response.meta["next"]
            }

    def file_handler(self):
        """
        Overrides default behavior to handle chunks of files
        @return: None
        """
        # Run as long as we have files to delete
        while self.task_active.is_set():
            try:
                # Get file non-blocking way, otherwise thread will hang forever
                file = self.handle_queue.get_nowait()
            except Empty:
                file = None

            if file:
                # If we have a file/folder to delete we add it to the deletion list
                local_path = "{}/{}".format(self.data_owner.working_dir, file["fullpath"])
                file["local_path"] = local_path
                if file.get("type", "files") == "files":
                    self.files_to_delete.append(file["fullpath"])
                elif file.get("type", "files") == "folders":
                    self.folders_to_delete.append(file["fullpath"])

            try:
                # Try to delete a chunk of files/folders
                if len(self.files_to_delete) + len(self.folders_to_delete) >= self.chunk_size or \
                   len(self.files_to_delete) + len(self.folders_to_delete) >= (self.total_files - self.handled_files):
                    self._handle_file_function()
            except Exception as e:
                # Should not be possible to enter here, safeguard against deadlock
                print("An unhandled exception in downloader thread has occurred:")
                print(e)
                with self.progress_lock:
                    self.handled_files += len(self.files_to_delete) + len(self.folders_to_delete)
                    self.files_to_delete = []
                    self.folders_to_delete = []

            if file is None:
                time.sleep(0.1)

    def _handle_file_function(self, **kwargs):
        """
        Function to that collect 1000 files that need to be deleted and then delete them from remote
        @param local_path: File location locally
        @param kwargs: Needs to be fullpath of the file
        @return: None
        """

        with self.progress_lock:
            try:
                self.data_owner._proxy.call_api(
                    urljoin(self.data_owner._route, "commits", self.commit, "delete_files"),
                    http_method=HTTP.POST,
                    payload={"files": self.files_to_delete, "mode": "files"}
                )
                self.data_owner._proxy.call_api(
                    urljoin(self.data_owner._route, "commits", self.commit, "delete_files"),
                    http_method=HTTP.POST,
                    payload={"files": self.folders_to_delete, "mode": "folders"}
                )
            finally:
                for f in self.files_to_delete:
                    self.handle_queue.task_done()
                    self.progress_queue.put(f)
                for f in self.folders_to_delete:
                    self.handle_queue.task_done()
                    self.progress_queue.put(f)
                self.files_to_delete = []
                self.folders_to_delete = []



from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
import pprint


class EndpointRule(DynamicAttributes):

    available_attributes = {
        "is_active": str,
        "title": str,
        "description": dict,
        "frequency": dict,
        "action": int,
        "severity": list,
        "min_events": str,
        "created_at": dict,
        "metric": str,
        "threshold": str,
        "operation": str,
    }

    def __init__(self, context, slug, attributes):
        self.slug = slug
        self._context = context
        self._attributes = attributes or {}

    def __str__(self):
        return pprint.pformat(self._attributes)



from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config import error_messages, routes


# The organization client is for internal use and is unsupported
class OrganizationsClient:
    def __init__(self, domain, token):
        self._domain = domain
        self._token = token
        self._proxy = Proxy(domain=domain, token=token)

    def get(self, slug):
        """
        Retrieves an organization by the given slug
        @param slug: The slug of the requested organization
        @return: Organization object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.ORGANIZATION_GET_FAULTY_SLUG)

        org_api = routes.ORGANIZATION_BASE.format(slug)

        response = self._proxy.call_api(
            route=org_api,
            http_method=HTTP.GET,
        )
        return response.attributes

    def create(self, name):
        """
        Creates a new organization with the provided name
        @param name: The name of the new organization
        @return: True if the organization was successfully created
        """
        if not name or not isinstance(name, str):
            raise CnvrgArgumentsError(error_messages.ORGANIZATION_CREATE_FAULTY_NAME)

        response = self._proxy.call_api(
            route=routes.ORGANIZATION_CREATE,
            http_method=HTTP.POST,
            payload={"title": name}
        )
        return response.attributes



from cnvrgv2.config import routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.images.image import Image
from cnvrgv2.modules.registries.utils import validate_registry_params
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.json_api_format import JAF


class Registry(DynamicAttributes):
    available_attributes = {
        "url": str,
        "title": str,
        "private": bool,
        "username": str,
        "registry_type": str,
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current project
        if slug:
            self._context.set_scope(SCOPE.REGISTRY, slug)

        scope = self._context.get_scope(SCOPE.REGISTRY)

        self.slug = scope["registry"]
        self._proxy = Proxy(context=self._context)
        self._route = routes.REGISTRY_BASE.format(scope["organization"], self.slug)
        self._attributes = attributes or {}

    @property
    def images(self):
        """
        Returns the registry images
        @return: [List] of Image objects
        """

        # TODO: convert to pagination
        response = self._proxy.call_api(route="{}/images".format(self._route), http_method=HTTP.GET)
        items = response.items

        images = []
        for item in items:
            slug = item.attributes["slug"]
            object_instance = Image(context=self._context, slug=slug, attributes=item.attributes)
            images.append(object_instance)

        return images

    def update(self, title=None, url=None, username=None, password=None):
        """
        Updates the current registry
        @param title: [String] The registry name
        @param url: [String] The registry url
        @param username: [String] The username with which to connect to the registry
        @param password: [String] The password with which to connect to the registry
        @return: Registry object
        """

        validate_registry_params(title=title, url=url, username=username, password=password)

        # Take only the attributes we want to update (remove None values)
        attributes = {"title": title, "url": url, "username": username, "password": password}
        attributes = {k: v for k, v in attributes.items() if v is not None}

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type="registry", attributes=attributes)
        )

        self._attributes = response.attributes
        return self

    def delete(self):
        """
        Deletes the current image
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)



import json
from collections import OrderedDict

import requests

from cnvrgv2._version import __version__
from cnvrgv2.config import Config
from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgError, CnvrgHttpError, ExceptionsConfig
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin


class HTTP:
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    OPTIONS = "OPTIONS"


class Proxy:
    def __init__(self, context=None, domain=None, token=None, is_capi=True):
        # Those are using for User/Org clients that do not have context
        self._domain = domain
        self._token = token
        self._is_capi = is_capi

        if context:
            self._domain = context.domain
            self._token = context.token
            self._is_capi = context.is_capi

    def call_api(self, route, http_method, payload=None, files_list=None, headers=None):
        """
        The main function controlling access to the Cnvrg API
        @param route: the partial url of the api resource to access
        @param http_method: get/post/put/delete
        @param payload: params to send along with the api request
        @param headers: headers to add to the http request. (application/json cont-type is sent by default)
        @param files_list: list of tuples in the following form: (file_name, file_path)
        @raise ValueError: if http_method is not legal
        @return: Response object
        """
        try:
            is_file = files_list is not None
            full_headers = self._build_headers(headers, is_file)
            full_url = urljoin(self._domain, "api", route)

            cookies = {}
            response = self._http_method_switch(http_method, full_url, payload, full_headers, cookies, files_list)
            return response

        except Exception as e:
            if ExceptionsConfig.SUPPRESS_EXCEPTIONS:
                # Return valid but empty response.
                return JAF({})
            else:
                raise e

    def _http_method_switch(self, http_method, url, payload, headers, cookies, files_list=None):
        """
        Executes the correct HTTP request based on the input parameters
        @param http_method: The https method to use
        @param url: The target's API url
        @param payload: The request's body json
        @param headers: Headers to attach to the request
        @param cookies: Cookies to attach to the request
        @param files_list: list of tuples in the following form: (file_name, file_path)
        @return: The response as a JAF object
        """
        check_certificate = Config().check_certificate or False
        response = None

        if http_method == HTTP.GET:
            response = requests.get(url, params=payload, headers=headers, cookies=cookies, verify=check_certificate)
        elif http_method == HTTP.POST and files_list is not None:
            files = []
            for name, path in files_list:
                files.append((name, (path, open(path, "rb"))))
            files.append(("data", (None, json.dumps(payload["data"]), "application/json")))
            response = requests.post(url, headers=headers, cookies=cookies, files=files)
        elif http_method == HTTP.POST:
            response = requests.post(url, json=payload, headers=headers, cookies=cookies, verify=check_certificate)
        elif http_method == HTTP.PUT:
            response = requests.put(url, json=payload, headers=headers, cookies=cookies, verify=check_certificate)
        elif http_method == HTTP.DELETE:
            response = requests.delete(url, json=payload, headers=headers, cookies=cookies, verify=check_certificate)
        return self._parse_response(response)

    def _build_headers(self, headers, is_file):
        """
        Builds the header object, composed from custom headers and mandatory headers
        @param headers: Custom headers to add to the request
        @return: A comprehensive headers object
        """
        if headers is None:
            headers = {}

        # Copy the headers so we wont change the original input
        full_headers = dict(headers)

        # If it's a file, we'll let requests lib complete the Content-Type (multipart/form-data)
        if not is_file:
            full_headers['Content-Type'] = "application/json"
        full_headers['User-Agent'] = "cnvrg/{version}".format(version=__version__)
        if self._is_capi:
            full_headers['Authorization'] = "CAPI {capi}".format(capi=self._token)
        else:
            full_headers['Auth-Token'] = self._token

        full_headers['Source'] = "sdk_v2"
        return full_headers

    def _parse_response(self, response):
        """
        This function handles the response from the api, decodes the json and returns the appropriate object
        @param response: Response object to handle
        @raise HttpError: if got error from the server
        @return: None or a json object
        """
        if response.status_code == requests.codes.ok:
            try:
                # Convert to ordered dict in order to keep the json ordering from the server
                response_decoded = response.json(object_pairs_hook=OrderedDict)
                return JAF(response=response_decoded)
            except Exception:
                raise CnvrgError(error_messages.PROXY_EMPTY_RESPONSE)

        elif response.status_code == requests.codes.no_content:
            return
        elif response.status_code == requests.codes.unauthorized:
            raise CnvrgHttpError(response.status_code, error_messages.PROXY_UNAUTH_ERROR)
        elif response.status_code == requests.codes.not_found:
            raise CnvrgHttpError(response.status_code, error_messages.PROXY_NOT_FOUND_ERROR)
        elif response.status_code >= 500:
            raise CnvrgHttpError(response.status_code, error_messages.PROXY_HTTP_ERROR)
        else:
            error_string = self._parse_errors(response.json())
            raise CnvrgHttpError(response.status_code, error_string)

    def _parse_errors(self, response):
        """
        Parse server errors in the format: { errors: [{ "title: "...", "detail": "..." }] }
        @param response: Error response from the server
        @return: String representing the errors
        """
        error_string = ""
        errors = response["errors"]
        for error in errors:
            error_string += error.get("title", "") + ": " + error.get("detail", "")

        return error_string



import os
import sys
import click
import codecs
import locale

from cnvrgv2.cli.modules.config import config
from cnvrgv2.cli.modules.ssh.ssh import ssh_group
from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.modules.users import users
from cnvrgv2.cli.logger.logger import CnvrgLogger
from cnvrgv2.cli.modules.project.project import project_group
from cnvrgv2.cli.modules.libraries.library import library_group
from cnvrgv2.cli.modules.dataset.dataset import dataset_group
from cnvrgv2.cli.modules.project.experiment import experiment_group
from cnvrgv2.cli.modules.project.workspace import workspace_group
from cnvrgv2.cli.modules.project.endpoint import endpoint_group
from cnvrgv2.cli.modules.images.images import image_group
from cnvrgv2.cli.modules.registries.registries import registry_group

"""
To avoid issues reading unicode chars from stdin or writing to stdout, we need to ensure that the
python3 runtime is correctly configured, if not, we try to force to utf-8.
"""
if codecs.lookup(locale.getpreferredencoding()).name == 'ascii':
    utf_8_locales = set([loc for loc in locale.locale_alias.values() if loc.lower().endswith((".utf-8", ".utf8"))])
    english_locales = [loc for loc in utf_8_locales if loc.lower().startswith('en_us')]

    if len(english_locales) > 0:
        os.environ['LANG'] = english_locales[0]
        os.environ['LC_ALL'] = english_locales[0]

    if codecs.lookup(locale.getpreferredencoding()).name == 'ascii':
        print("Could not set locale variable automatically.")
        print("Please set the LANG and LC_ALL env variables to a proper value before calling this script.")
        print("The following suitable locales were discovered:")
        print(', '.join('{}'.format(k) for k in utf_8_locales))
        sys.exit(-1)


@click.group()
def entry_point():
    pass


def safe_entry_point():
    try:
        entry_point()
    except Exception as e:
        click.echo(messages.CLI_UNEXPECTED_ERROR.format(str(e)), err=True)
        exit(1)


entry_point.add_command(config)
entry_point.add_command(ssh_group)
entry_point.add_command(image_group)
entry_point.add_command(users.login)
entry_point.add_command(users.logout)
entry_point.add_command(project_group)
entry_point.add_command(library_group)
entry_point.add_command(dataset_group)
entry_point.add_command(endpoint_group)
entry_point.add_command(registry_group)
entry_point.add_command(workspace_group)
entry_point.add_command(experiment_group)
entry_point.add_command(CnvrgLogger.set_logs_keep_duration)



from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.config import routes, error_messages
from cnvrgv2.utils.validators import validate_email, validate_user_role
from cnvrgv2 import cnvrg
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config.error_messages import FAULTY_VALUE


class MembersClient:
    def __init__(self, organization):
        if type(organization) is not cnvrg.Cnvrg:
            raise CnvrgArgumentsError(error_messages.ARGUMENT_BAD_TYPE.format(cnvrg.Cnvrg, type(organization)))
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.ORGANIZATION_MEMBERS.format(scope["organization"])

    def add(self, email, role):
        """
        Add a new member to the organization
        @param email: User email
        @param role: Role to add
        """

        if not validate_user_role(role):
            raise CnvrgArgumentsError(FAULTY_VALUE.format(role))

        if not validate_email(email):
            raise CnvrgArgumentsError(FAULTY_VALUE.format(email))

        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="users", attributes={"email": email, "role": role})
        )

    def revoke(self, email):
        """
        Revoke user access to organization
        @param email: User's email
        """
        if not validate_email(email):
            raise CnvrgArgumentsError(FAULTY_VALUE.format(email))

        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.DELETE,
            payload=JAF.serialize(type="users", attributes={"email": email})
        )
